<?php
// imports
if (!defined('NO_IMPORTS')) {
	require_once(WCF_DIR.'lib/system/exception/SystemException.class.php');
	require_once(WCF_DIR.'lib/system/io/File.class.php');
}

/**
 * CacheHandler holds all registered cache resources.
 * 
 * @package	com.woltlab.wcf.system.cache
 * @author	Marcel Werk
 * @copyright	2001-2007 WoltLab GmbH
 * @license	GNU Lesser General Public License <http://opensource.org/licenses/lgpl-license.php>
 */
class CacheHandler {
	/**
	 * Loaded cache.
	 * 
	 * @var array
	 */
	protected $cache;
	
	/**
	 * Registered cache resources.
	 * 
	 * @var array
	 */
	protected $cacheResources = array();
	
	/**
	 * List of loaded resources.
	 * 
	 * @var array
	 */
	protected $loaded = array();
	
	/**
	 * Registers a new cache resource.
	 * 
	 * @param	string		$cache		name of this resource
	 * @param	string		$file		data file for this resource
	 * @param	string		$classFile
	 * @param	integer		$minLifetime
	 * @param	integer		$maxLifetime
	 */
	public function addResource($cache, $file, $classFile, $minLifetime = 0, $maxLifetime = 0) {
		$className = StringUtil::getClassName($classFile);
		
		$this->cacheResources[$cache] = array(
			'cache' => $cache,
			'file' => $file, 
			'className' => $className, 
			'classFile' => $classFile,
			'minLifetime' => $minLifetime,
			'maxLifetime' => $maxLifetime
		);
	}
	
	/**
	 * Deletes a registered cache resource.
	 * 
	 * @param 	string		$cache
	 * @param 	boolean		$ignoreLifetime
	 */
	public function clearResource($cache, $ignoreLifetime = false) {
		if (!isset($this->cacheResources[$cache])) {
			throw new SystemException("cache resource '".$cache."' does not exist", 11005);
		}
		
		if (file_exists($this->cacheResources[$cache]['file'])) {
			if ($ignoreLifetime || ($this->cacheResources[$cache]['minLifetime'] == 0 || (TIME_NOW - filemtime($this->cacheResources[$cache]['file'])) >= $this->cacheResources[$cache]['minLifetime'])) {
				// delete cache file
				@unlink($this->cacheResources[$cache]['file']);
				
				// reset open cache
				if (isset($this->cache[$cache])) unset($this->cache[$cache]);
				if (isset($this->loaded[$this->cacheResources[$cache]['file']])) unset($this->loaded[$this->cacheResources[$cache]['file']]);
			}
		}
	}
	
	/**
	 * Determines wheater the cache needs to be rebuilt or not.
	 *
	 * @param 	array 		$cacheResource
	 * @return 	boolean 	$needRebuilt
	 */
	protected function needRebuilt($cacheResource) {
		// cache does not exist
		if (!file_exists($cacheResource['file'])) {
			return true;	
		}
		
		// cache is empty
		if (!@filesize($cacheResource['file'])) {
			return true;	
		}
		
		// cache resource was marked as obsolete
		if (($mtime = filemtime($cacheResource['file'])) <= 1) {
			return true;	
		}
		
		// cache resource has a delay marking
		if ($cacheResource['minLifetime'] > 0 && (TIME_NOW - $mtime) < $cacheResource['minLifetime']) {
			return false;
		}
		
		// maxlifetime expired
		if ($cacheResource['maxLifetime'] > 0 && (TIME_NOW - $mtime) > $cacheResource['maxLifetime']) {
			return true;	
		}
		
		// do not rebuild cache
		return false;
	}
	
	/**
	 * Marks cached files as obsolete.
	 *
	 * @param 	string 		$directory
	 * @param 	string 		$filepattern
	 * @param 	boolean		$forceDelete
	 */
	public function clear($directory, $filepattern, $forceDelete = false) {
		$filepattern = str_replace('*', '.*', str_replace('.', '\.', $filepattern));
		if (substr($directory, -1) != '/') {
			$directory .= '/';	
		}
		
		if (@file_exists($directory)) {
			$dirh = opendir($directory);
			while ($filename = readdir($dirh)) {
				if ($filename != '.' && $filename != '..' && preg_match('%^'.$filepattern.'$%i', $filename)) {
					if ($forceDelete || !@touch($directory.$filename, 1)) {
						@unlink($directory.$filename);
					}
				}
			}
			closedir($dirh);
		}
	}
	
	/**
	 * Returns a cached variable.
	 *
	 * @param 	string 		$cache
	 * @param 	string 		$variable
	 * @return 	mixed 		$value
	 */
	public function get($cache, $variable = '') {
		if (!isset($this->cache[$cache]) && isset($this->cacheResources[$cache])) {
			// load cache
			$this->load($this->cacheResources[$cache]);
		}
		
		if (!isset($this->cache[$cache])) {
			// try to rebuild and reload the cache
			$this->rebuild($this->cacheResources[$cache], true);
			$this->load($this->cacheResources[$cache], true);
			
			if (!isset($this->cache[$cache])) {
				throw new SystemException("cache resource '".$cache."' does not exist", 11005);
			}
		}
		
		if (!empty($variable)) {
			if (!isset($this->cache[$cache][$variable])) {
				throw new SystemException("variable '".$variable."' does not exist in cache resource '".$cache."'", 11008);
			}
			
			return $this->cache[$cache][$variable];
		}
		else {
			return $this->cache[$cache];
		}
	}
	
	/**
	 * Loads a cached resource.
	 * 
	 * @param 	array 		$cacheResource
	 * @param	boolean		$reload
	 */
	public function load($cacheResource, $reload = false) {
		if ($reload || !isset($this->loaded[$cacheResource['file']]) || !$this->loaded[$cacheResource['file']]) {
			if ($this->rebuild($cacheResource)) {
				// now include the cache
				try {
					$this->loadCacheFile($cacheResource);
				}
				catch (Exception $e) {
					// include failed; force rebuild and try again
					$this->rebuild($cacheResource, true);
					$this->loadCacheFile($cacheResource);
				}
			}
			$this->loaded[$cacheResource['file']] = true;
		}
	}
	
	/**
	 * Loads the file of a cached resource.
	 * 
	 * @param 	array 		$cacheResource
	 */
	protected function loadCacheFile($cacheResource) {
		// get file contents
		$contents = file_get_contents($cacheResource['file']);
		
		// find first newline
		$position = strpos($contents, "\n");
		if ($position === false) throw new SystemException("Unable to load cache resource '".$cacheResource['cache']."'");
		
		// cut contents
		$contents = substr($contents, $position + 1);
		
		// unserialize
		$this->cache[$cacheResource['cache']] = @unserialize($contents);
		if ($this->cache[$cacheResource['cache']] === false) throw new SystemException("Unable to load cache resource '".$cacheResource['cache']."'");
	}
	
	/**
	 * Rebuilds the cache.
	 * 
	 * @param 	array 		$cacheResource
	 * @param	boolean		$forceRebuild
	 * @return 	boolean 	result
	 */
	public function rebuild($cacheResource, $forceRebuild = false) {
		// determine whether the cache needs to be rebuild or not
		if (!$forceRebuild && !$this->needRebuilt($cacheResource)) {
			return true;
		}
		
		// include cache resource class
		if (!file_exists($cacheResource['classFile'])) {
			throw new SystemException("Unable to find class file '".$cacheResource['classFile']."'", 11000);
		}
		require_once($cacheResource['classFile']);
		
		// instance cache class
		if (!class_exists($cacheResource['className'])) {
			throw new SystemException("Unable to find class '".$cacheResource['className']."'", 11001);
		}
		$cacheBuilder = new $cacheResource['className'];
		
		// write cache
		$targetFile = new File($cacheResource['file']);
		$targetFile->write("<?php exit; /* cache: ".$cacheResource['cache']." (generated at ".gmdate('r').") DO NOT EDIT THIS FILE */ ?>\n");
		$targetFile->write(serialize($cacheBuilder->getData($cacheResource)));
		$targetFile->close();
		
		// reload cache
		if (isset($this->loaded[$cacheResource['file']]) && $this->loaded[$cacheResource['file']]) {
			$this->load($cacheResource, true);	
		}
		
		return true;
	}
}
?>