<?php
// imports
if (!defined('NO_IMPORTS')) {
	require_once(WCF_DIR.'lib/system/exception/SystemException.class.php');
	require_once(WCF_DIR.'lib/system/io/File.class.php');
	require_once(WCF_DIR.'lib/system/template/Template.class.php');
	require_once(WCF_DIR.'lib/system/template/TemplatePluginPrefilter.class.php');
	require_once(WCF_DIR.'lib/system/template/TemplatePluginCompiler.class.php');
}

/**
 * TemplateCompiler compiles template source in valid php code.
 * 
 * @package	com.woltlab.wcf.system.template
 * @author	Marcel Werk
 * @copyright	2001-2007 WoltLab GmbH
 * @license	GNU Lesser General Public License <http://opensource.org/licenses/lgpl-license.php>
 */
class TemplateCompiler {
	protected 	$constantPattern, $doubleQuotePattern, $singleQuotePattern,
			$quotePattern, $numericPattern, $simpleVarPattern,
			$dotAccessPattern, $bracketAccessPattern, $varWithoutObjectAccessPattern,
			$objectMethodParamPattern, $objectAccessPattern, $varPattern,
			$varAndMorePattern, $mathOperatorsPattern, $varWithMathPattern,
			$modifierPattern, $varWithModifierPattern, $outputPattern,
			$escapedPattern, $template, $currentTemplate, 
			$currentLineNo;
			
	protected 	$modifiers = array(), $autoloadPlugins = array(), $tagStack = array(),
			$compilerPlugins = array(), $captureStack = array();
	
	protected 	$leftDelimiter = '{', $rightDelimiter = '}';
	protected	$unknownPHPFunctions = array('isset', 'unset', 'empty');
	
	/**
	 * Creates a new TemplateCompiler.
	 * 
	 * @param	Template		$template
	 */
	public function __construct(Template $template) {
		$this->template = $template;
		$this->buildPattern();
	}
	
	/**
	 * Compiles the source of a template.
	 * 
	 * @param	string		$templateName
	 * @param	string		$sourceContent
	 * @param	string		$compiledFilename
	 */
	public function compile($templateName, $sourceContent, $compiledFilename) {
		// reset vars
		$this->autoloadPlugins = $this->tagStack = $this->stringStack = $this->literalStack = array();
		$this->currentTemplate = $templateName;
		$this->currentLineNo = 1;
		
		// quote left and right delimiter for use in
		// regular expressions
		$ldq = preg_quote($this->leftDelimiter, '~').'(?=\S)';
		$rdq = '(?<=\S)'.preg_quote($this->rightDelimiter, '~');
		
		// apply prefilters
		foreach ($this->template->getPrefilters() as $prefilter) {
			if (!is_object($prefilter)) {
				$filename = $this->template->getPluginFilename('prefilter', $prefilter);
				if (!file_exists($filename)) {
					throw new SystemException('unable to find class file '.$filename, 11000);
				}
				require_once($filename);
				$className = 'TemplatePluginPrefilter'.StringUtil::firstCharToUpperCase(StringUtil::toLowerCase($prefilter));
				
				if (!class_exists($className)) {
					throw new SystemException($this->formatSyntaxError('unable to find prefilter class '.$className, $this->currentTemplate), 11001);
				}
				$prefilter = new $className();
			}
			
			if ($prefilter instanceof TemplatePluginPrefilter) {
				$sourceContent = $prefilter->execute($sourceContent, $this);
			}
			else {
				throw new SystemException($this->formatSyntaxError("Prefilter '".$prefilter."' does not implement the interface 'TemplatePluginPrefilter'", $this->currentTemplate), 11010);
			}
		}
		
		// replace all {literal} Tags with unique hash values
		$sourceContent = preg_replace("~{$ldq}literal{$rdq}(.*?){$ldq}/literal{$rdq}~es", 'StringStack::pushToStringStack(StringUtil::unescape(\'$1\'), \'literal\')', $sourceContent);
		
		// handle <?php tags
		if (StringUtil::indexOf($sourceContent, '<?') !== false) {
			$sourceContent = StringUtil::replace('<?php', '@@PHP_START_TAG@@', $sourceContent);
			$sourceContent = StringUtil::replace('<?', '@@PHP_SHORT_START_TAG@@', $sourceContent);
			$sourceContent = StringUtil::replace('?>', '@@PHP_END_TAG@@', $sourceContent);
			$sourceContent = StringUtil::replace('@@PHP_END_TAG@@', "<?php echo '?>'; ?>\n", $sourceContent);
			$sourceContent = StringUtil::replace('@@PHP_SHORT_START_TAG@@', "<?php echo '<?'; ?>\n", $sourceContent);
			$sourceContent = StringUtil::replace('@@PHP_START_TAG@@', "<?php echo '<?php'; ?>\n", $sourceContent);
		}
		
		// remove comments
		$sourceContent = preg_replace("~{$ldq}\*.*?\*{$rdq}~s", '', $sourceContent);
		
		// match all template tags
		$matches = array();
		preg_match_all("~{$ldq}(.*?){$rdq}~s", $sourceContent, $matches);
		$templateTags = $matches[1];
		
		// Split content by template tags to obtain non-template content
		$textBlocks = preg_split("~{$ldq}.*?{$rdq}~s", $sourceContent);
		
		// compile the template tags into php-code
		$compiledTags = array();
		for ($i = 0, $j = count($templateTags); $i < $j; $i++) {
			$this->currentLineNo += StringUtil::countSubstring($textBlocks[$i], "\n");
			$compiledTags[] = $this->compileTag($templateTags[$i]);
			$this->currentLineNo += StringUtil::countSubstring($templateTags[$i], "\n");
		}
		
		// throw error messages for unclosed tags
		if (count($this->tagStack) > 0) {
			foreach ($this->tagStack as $tagStack) {
				throw new SystemException($this->formatSyntaxError('unclosed tag {'.$tagStack[0].'}', $this->currentTemplate, $tagStack[1]), 12002);
			}
			return false;
		}
		
		$compiledContent = '';
		// Interleave the compiled contents and text blocks to get the final result.
		for ($i = 0, $j = count($compiledTags); $i < $j; $i++) {
			if ($compiledTags[$i] == '') {
				// tag result empty, remove first newline from following text block
				$textBlocks[$i + 1] = preg_replace('%^(\r\n|\r|\n)%', '', $textBlocks[$i + 1]);
			}
			$compiledContent .= $textBlocks[$i].$compiledTags[$i];
		}
		$compiledContent .= $textBlocks[$i];
		$compiledContent = chop($compiledContent);
		
		// INSERT POSTFILTERS HERE!
		
		// reinsert {literal} Tags
		$compiledContent = StringStack::reinsertStrings($compiledContent, 'literal');
		
		// build fileheader
		$compiledHeader = "<?php\n/**\n* WoltLab Community Framework\n* Template: ".$templateName."\n* Compiled at: ".gmdate('r')."\n* \n* DO NOT EDIT THIS FILE\n*/\n?>\n";
		
		// include Plugins
		$compiledAutoloadPlugins = '';
		if (count($this->autoloadPlugins) > 0) {
			$compiledAutoloadPlugins = "<?php\n";
			foreach ($this->autoloadPlugins as $className => $fileName) {
				$compiledAutoloadPlugins .= "if (!isset(\$this->pluginObjects['$className'])) {\n";
				$compiledAutoloadPlugins .= "require_once(WCF_DIR.'".StringUtil::replace(WCF_DIR, '', $fileName)."');\n";
				$compiledAutoloadPlugins .= "\$this->pluginObjects['$className'] = new $className;\n";
				$compiledAutoloadPlugins .= "}\n";
			}
			$compiledAutoloadPlugins .= "?>";
		}
		
		// write compiled to file
		$file = new File($compiledFilename);
		$file->write($compiledHeader.$compiledAutoloadPlugins.$compiledContent);
		$file->close();
	}
	
	/**
	 * Compiles a template tag.
	 * 
	 * @param	string		$tag
	 */
	protected function compileTag($tag) {
		if (preg_match('~^'.$this->outputPattern.'$~s', $tag)) {
			// variable output
			return $this->compileOutputTag($tag);
		}
		
		$match = array();
		// replace 'else if' with 'elseif'
		$tag = preg_replace('~^else\s+if(?=\s)~i', 'elseif', $tag);
		
		if (preg_match('~^(/?\w+)~', $tag, $match)) {
			// build in function or plugin
			$tagCommand = $match[1];
			$tagArgs = StringUtil::substring($tag, StringUtil::length($tagCommand));
			
			switch ($tagCommand) {
				case 'if':
					$this->pushTag('if');
					return $this->compileIfTag($tagArgs);
					
				case 'elseif':
				case 'else if':
					list($openTag) = end($this->tagStack);
					if ($openTag != 'if' && $openTag != 'elseif') {
						throw new SystemException($this->formatSyntaxError('unxepected {elseif}', $this->currentTemplate, $this->currentLineNo));
					}
					else if ($openTag == 'if') {
						$this->pushTag('elseif');
					}
					return $this->compileIfTag($tagArgs, true);
					
				case 'else':
					list($openTag) = end($this->tagStack);
					if ($openTag != 'if' && $openTag != 'elseif') {
						throw new SystemException($this->formatSyntaxError('unexpected {else}', $this->currentTemplate, $this->currentLineNo));
					}
					else {
						$this->pushTag('else');
						return '<?php } else { ?>';
					}
					
				case '/if':
					list($openTag) = end($this->tagStack);
					if ($openTag != 'if' && $openTag != 'elseif' && $openTag != 'else') {
						throw new SystemException($this->formatSyntaxError('unexpected {/if}', $this->currentTemplate, $this->currentLineNo));
					}
					else {
						$this->popTag('if');
					}
					return '<?php } ?>';
				
				case 'include':
					return $this->compileIncludeTag($tagArgs);
					
				case 'foreach':
					$this->pushTag('foreach');
					return $this->compileForeachTag($tagArgs);
					
				case '/foreach':
					$this->popTag('foreach');
					return "<?php } ?>";
					
				case 'section':
					$this->pushTag('section');
					return $this->compileSectionTag($tagArgs);
			
				case '/section':
					$this->popTag('section');
					return "<?php } } } ?>";
					
				case 'capture':
					$this->pushTag('capture');
					return $this->compileCaptureTag(true, $tagArgs);
			
				case '/capture':
					$this->popTag('capture');
					return $this->compileCaptureTag(false);
				
				default:
					// 1) compiler functions first
					if ($phpCode = $this->compileCompilerPlugin($tagCommand, $tagArgs)) {
						return $phpCode;
					}
					// 2) block functions
					if ($phpCode = $this->compileBlockPlugin($tagCommand, $tagArgs)) {
						return $phpCode;
					}
					// 3) functions
					if ($phpCode = $this->compileFunctionPlugin($tagCommand, $tagArgs)) {
						return $phpCode;
					}
			}
			
		}
		
		throw new SystemException($this->formatSyntaxError('unknown tag {'.$tag.'}', $this->currentTemplate, $this->currentLineNo), 12003);
	}
	
	/**
	 * Compiles a function plugin.
	 *
	 * @param 	string 		$tagCommand
	 * @param 	string 		$tagArgs
	 * @return 	mixed				false, if the plugin does not exist
	 * 						otherwise the php output of the plugin
	 */
	protected function compileFunctionPlugin($tagCommand, $tagArgs) {
		// if necessary load plugin from plugin-dir
		if (!file_exists($pluginFilename = $this->template->getPluginFilename('function', $tagCommand))) {
			return false;
		}
		
		$className = 'TemplatePluginFunction'.StringUtil::firstCharToUpperCase(StringUtil::toLowerCase($tagCommand));
		$this->autoloadPlugins[$className] = $pluginFilename;
		
		$tagArgs = $this->makeArgString($this->parseTagArgs($tagArgs, $tagCommand));

		return "<?php echo \$this->pluginObjects['$className']->execute(array($tagArgs), \$this); ?>";
	}
	
	/**
	 * Compiles a block plugin.
	 *
	 * @param 	string 		$tagCommand
	 * @param 	string 		$tagArgs
	 * @return 	mixed				false, if the plugin does not exist
	 * 						otherwise the php output of the plugin
	 */
	protected function compileBlockPlugin($tagCommand, $tagArgs) {
		// check wheater this is the start ({block}) or the
		// end tag ({/block})
		if (substr($tagCommand, 0, 1) == '/') {
			$tagCommand = substr($tagCommand, 1);
			$startTag = false;
		}
		else {
			$startTag = true;
		}

		// if necessary load plugin from plugin-dir
		if (!file_exists($pluginFilename = $this->template->getPluginFilename('block', $tagCommand))) {
			return false;
		}
			
		$className = 'TemplatePluginBlock'.StringUtil::firstCharToUpperCase(StringUtil::toLowerCase($tagCommand));
		$this->autoloadPlugins[$className] = $pluginFilename;
		
		if ($startTag) {
			$this->pushTag($tagCommand);

			$tagArgs = $this->makeArgString($this->parseTagArgs($tagArgs, $tagCommand));

			$phpCode = "<?php \$this->tagStack[] = array('$tagCommand', array($tagArgs));\n";
			$phpCode .= "\$this->pluginObjects['".$className."']->init(\$this->tagStack[count(\$this->tagStack) - 1][1], \$this);\n";
			$phpCode .= "while (\$this->pluginObjects['".$className."']->next(\$this)) { ob_start(); ?>";
		}
		else {
			$this->popTag($tagCommand);
			$phpCode = "<?php \$blockContent = ob_get_contents();\n";
			$phpCode .= "ob_end_clean();\n";
			$phpCode .= "echo \$this->pluginObjects['$className']->execute(\$this->tagStack[count(\$this->tagStack) - 1][1], \$blockContent, \$this); }\n";
			$phpCode .= "array_pop(\$this->tagStack);\n";
			$phpCode .= "unset(\$blockContent, \$blockRepeat); ?>";
		}
		
		return $phpCode;
	}
	
	
	/**
	 * Compiles a compiler function/block.
	 *
	 * @param 	string 		$tagCommand
	 * @param 	string 		$tagArgs
	 * @return 	mixed				false, if the plugin does not exist
	 * 						otherwise the php output of the plugin
	 */
	protected function compileCompilerPlugin($tagCommand, $tagArgs) {
		// check wheater this is the start ({block}) or the
		// end tag ({/block})
		if (substr($tagCommand, 0, 1) == '/') {
			$tagCommand = substr($tagCommand, 1);
			$startTag = false;
		}
		else {
			$startTag = true;
		}

		$className = 'TemplatePluginCompiler'.StringUtil::firstCharToUpperCase(StringUtil::toLowerCase($tagCommand));
		// if necessary load plugin from plugin-dir
		if (!isset($this->compilerPlugins[$className])) {
			if (!file_exists($pluginFilename = $this->template->getPluginFilename('compiler', $tagCommand))) {
				return false;
			}
			
			include_once($pluginFilename);
			if (!class_exists($className)) {
				throw new SystemException($this->formatSyntaxError('unable fo find compiler plugin class '.$className, $this->currentTemplate), 11001);
			}
			
			$this->compilerPlugins[$className] = new $className();
			
			if (!($this->compilerPlugins[$className] instanceof TemplatePluginCompiler)) {
				throw new SystemException($this->formatSyntaxError("Compiler plugin '".$tagCommand."' does not implement the interface 'TemplatePluginCompiler'", $this->currentTemplate), 11010);
			}
		}
		
		// execute plugin
		if ($startTag) {
			$tagArgs = $this->parseTagArgs($tagArgs, $tagCommand);
			$phpCode = $this->compilerPlugins[$className]->executeStart($tagArgs, $this);
		}
		else {
			$phpCode = $this->compilerPlugins[$className]->executeEnd($this);
		}
		
		return $phpCode;
	}
		
	/**
	 * Compiles a capture tag.
	 *
	 * @param 	boolean 	$startTag
	 * @param 	string 		$captureTag
	 * @return 	string 				phpCode
	 */
	protected function compileCaptureTag($startTag, $captureTag = null) {
		if ($startTag) {
			$append = false;
			$args = $this->parseTagArgs($captureTag, 'capture');
			
			if (!isset($args['assign'])) {
				if (!isset($args['append'])) {
					throw new SystemException($this->formatSyntaxError("missing 'assign' attribute in capture tag", $this->currentTemplate, $this->currentLineNo), 12001);
				}
				else {
					$args['assign'] = $args['append'];
					$append = true;
				}
			}
			/*$args['assign'] = $this->dequote($args['assign']);
			if (StringUtil::substring($args['assign'], 0, 1) != '$') {
				$args['assign'] = '$'.$args['assign'];
				$args['assign'] = $this->compileVariableTag($args['assign']);
			}*/
			
			$this->captureStack[] = array('variable' => $args['assign'], 'append' => $append);
			return '<?php ob_start(); ?>';
		}
		else {
			$capture = array_pop($this->captureStack);
			return '<?php $this->'.($capture['append'] ? 'append' : 'assign').'('.$capture['variable'].', ob_get_contents()); ob_end_clean(); ?>';
		}
	}
	
	/**
	 * Compiles a section tag.
	 *
	 * @param 	string 		$sectionTag
	 * @return 	string 				phpCode
	 */
	protected function compileSectionTag($sectionTag) {
		$args = $this->parseTagArgs($sectionTag, 'section');

		// check arguments
		if (!isset($args['loop'])) {
			throw new SystemException($this->formatSyntaxError("missing 'loop' attribute in section tag", $this->currentTemplate, $this->currentLineNo), 12001);
		}
		if (!isset($args['name'])) {
			throw new SystemException($this->formatSyntaxError("missing 'name' attribute in section tag", $this->currentTemplate, $this->currentLineNo), 12001);
		}
		if (!isset($args['show'])) {
			$args['show'] = true;
		}
		
		// dequote name argument
		$args['name'] = $this->dequote($args['name']);
		$sectionName = $args['name'];
		if (StringUtil::substring($args['name'], 0, 1) != '$') {
			$args['name'] = '$'.$args['name'];
			$args['name'] = $this->compileVariableTag($args['name']);
		}

		$sectionProp = "\$section['$sectionName']";

		$phpCode = "<?php\n";
		$phpCode .= "if (".(StringUtil::substring($args['loop'], 0, 1) == '$' ? "isset(".$args['loop'].")" : "true").") {\n";
		$phpCode .= $sectionProp." = array();\n";
		$phpCode .= $sectionProp."['loop'] = (is_array(".$args['loop'].") ? count(".$args['loop'].") : max(0, (int)".$args['loop']."));\n";
		$phpCode .= $sectionProp."['show'] = ".$args['show'].";\n";
		if (!isset($args['step'])) {
			$phpCode .= $sectionProp."['step'] = 1;\n";
		}
		else {
			$phpCode .= $sectionProp."['step'] = ".$args['step'].";\n";
		}
		if (!isset($args['max'])) {
			$phpCode .= $sectionProp."['max'] = ".$sectionProp."['loop'];\n";
		}
		else {
			$phpCode .= $sectionProp."['max'] = (".$args['max']." < 0 ? ".$sectionProp."['loop'] : ".$args['max'].");\n";
		}
		if (!isset($args['start'])) {
			$phpCode .= $sectionProp."['start'] = (".$sectionProp."['step'] > 0 ? 0 : ".$sectionProp."['loop'] - 1);\n";
		}
		else {
			$phpCode .= $sectionProp."['start'] = ".$args['start'].";\n";
			$phpCode .= "if (".$sectionProp."['start'] < 0) {\n";
			$phpCode .= $sectionProp."['start'] = max(".$sectionProp."['step'] > 0 ? 0 : -1, ".$sectionProp."['loop'] + ".$sectionProp."['start']);\n}\n";
			$phpCode .= "else {\n";
			$phpCode .= $sectionProp."['start'] = min(".$sectionProp."['start'], ".$sectionProp."['step'] > 0 ? ".$sectionProp."['loop'] : ".$sectionProp."['loop'] - 1);\n}\n";
		}
		$phpCode .= "if (".$args['show'].") {\n";

		if (!isset($args['start']) && !isset($args['step']) && !isset($args['max'])) {
			$phpCode .= $sectionProp."['total'] = ".$sectionProp."['loop'];\n";
		} else {
			$phpCode .= $sectionProp."['total'] = min(ceil((".$sectionProp."['step'] > 0 ? ".$sectionProp."['loop'] - ".$sectionProp."['start'] : ".$sectionProp."['start'] + 1) / abs(".$sectionProp."['step'])), ".$sectionProp."['max']);\n";
		}
		$phpCode .= "if (".$sectionProp."['total'] == 0)\n";
		$phpCode .= $sectionProp."['show'] = false;\n";
		$phpCode .= "} else\n";
		$phpCode .= "".$sectionProp."['total'] = 0;\n";

		$phpCode .= "if (".$sectionProp."['show']) {\n";
		$phpCode .= "for (".$args['name']." = ".$sectionProp."['start'], ".$sectionProp."['iteration'] = 1;\n";
		$phpCode .= $sectionProp."['iteration'] <= ".$sectionProp."['total'];\n";
		$phpCode .= $args['name']." += ".$sectionProp."['step'], ".$sectionProp."['iteration']++) {\n";
		$phpCode .= $sectionProp."['rowNumber'] = ".$sectionProp."['iteration'];\n";
		$phpCode .= $sectionProp."['indexPrev'] = ".$args['name']." - ".$sectionProp."['step'];\n";
		$phpCode .= $sectionProp."['indexNext'] = ".$args['name']." + ".$sectionProp."['step'];\n";
		$phpCode .= $sectionProp."['first']      = (".$sectionProp."['iteration'] == 1);\n";
		$phpCode .= $sectionProp."['last']       = (".$sectionProp."['iteration'] == ".$sectionProp."['total']);\n";
		$phpCode .= "?>";

		return $phpCode;
	}
	
	/**
	 * Compiles a foreach tag.
	 *
	 * @param 	string 		$foreachTag
	 * @return 	string 				phpCode
	 */
	protected function compileForeachTag($foreachTag) {
		$args = $this->parseTagArgs($foreachTag, 'foreach');

		// check arguments
		if (!isset($args['from'])) {
			throw new SystemException($this->formatSyntaxError("missing 'from' attribute in foreach tag", $this->currentTemplate, $this->currentLineNo), 12001);
		}
		if (!isset($args['item'])) {
			throw new SystemException($this->formatSyntaxError("missing 'item' attribute in foreach tag", $this->currentTemplate, $this->currentLineNo), 12001);
		}

		// dequote item & key arguments
		$args['item'] = $this->dequote($args['item']);
		if (StringUtil::substring($args['item'], 0, 1) != '$') {
			$args['item'] = '$'.$args['item'];
			$args['item'] = $this->compileVariableTag($args['item']);
		}
		
		if (isset($args['key'])) {
			$args['key'] = $this->dequote($args['key']);
			if (StringUtil::substring($args['key'], 0, 1) != '$') {
				$args['key'] = '$'.$args['key'];
				$args['key'] = $this->compileVariableTag($args['key']);
			}
		}

		if (isset($args['key'])) {
			$phpCode = "<?php foreach (".$args['from']." as ".$args['key']." => ".$args['item'].") { ?>";
		}
		else {
			$phpCode = "<?php foreach (".$args['from']." as ".$args['item'].") { ?>";
		}
		return $phpCode;
	}
	
	/**
	 * Compiles an include tag.
	 *
	 * @param 	string 		$includeTag
	 * @return 	string 				phpCode
	 */
	protected function compileIncludeTag($includeTag) {
		$args = $this->parseTagArgs($includeTag, 'include');
		$append = false;
		
		// check arguments
		if (!isset($args['file'])) {
			throw new SystemException($this->formatSyntaxError("missing 'file' attribute in include tag", $this->currentTemplate, $this->currentLineNo), 12001);
		}

		// get filename
		$file = $args['file'];
		unset($args['file']);

		// special parameters
		$assignVar = false;
		if (isset($args['assign'])) {
			$assignVar = $args['assign'];
			unset($args['assign']);
		}
		
		if (isset($args['append'])) {
			$assignVar = $args['append'];
			$append = true;
			unset($args['append']);
		}
		
		$sandbox = true;
		if (isset($args['sandbox'])) {
			$sandbox = false;
			unset($args['sandbox']);
		}

		// make argument string
		$argString = $this->makeArgString($args);
		
		// build phpCode
		$phpCode = '<?php ';
		
		if ($assignVar !== false) {
			$phpCode .= "ob_start();\n";
		}
		
		$phpCode .= '$this->includeTemplate('.$file.', array('.$argString.'), '.($sandbox ? 1 : 0).');'."\n";
		
		if ($assignVar !== false) {
			$phpCode .= '$this->'.($append ? 'append' : 'assign').'('.$assignVar.', ob_get_contents()); ob_end_clean();'."\n";
		}
		
		$phpCode .= ' ?>';
		
		return $phpCode;
	}
	
	/**
	 * Parses an argument list and returns
	 * the keys and values in an associative array.
	 *
	 * @param 	string 		$tagArgs
	 * @param 	string		$tag
	 * @return 	array 		$tagArgs
	 */
	public function parseTagArgs($tagArgs, $tag) {
		// match valid tag
		if (!preg_match('~^(?:\s+\w+=(?:'.$this->varWithModifierPattern.'|\w+))*$~', $tagArgs)) {
			throw new SystemException($this->formatSyntaxError('syntax error in tag {'.$tag.'}', $this->currentTemplate, $this->currentLineNo), 12000);
		}
		
		// find all variables
		$matches = array();
		preg_match_all('~\s+(\w+)=('.$this->varWithModifierPattern.'|\w+)~', $tagArgs, $matches);
		$args = array();
		for ($i = 0, $j = count($matches[1]); $i < $j; $i++) {
			$name = $matches[1][$i];
			$value = $matches[2][$i];
			
			if (preg_match('~^\w+$~', $value) && !preg_match('~^[A-Z0-9_]+$~', $value) && !preg_match('~^\d+$~', $value) && $value != 'true' && $value != 'false') {
				$value = "'".$value."'";
			}
			else {
				$value = $this->compileVariableTag($value);
			}
			
			$args[$name] = $value;
		}
		
		return $args;
	}
	
	/**
	 * Takes an array created by TemplateCompiler::parseTagArgs()
	 * and creates a string.
	 *
	 * @param 	array 		$args
	 * @return 	string 		$args
	 */
	public static function makeArgString($args) {
		$argString = '';
		foreach	($args as $key => $val) {
			if ($argString != '') {
				$argString .= ', ';
			}
			$argString .= "'$key' => $val";
		}
		return $argString;
	}
	
	/**
	 * Formats a syntax error message.
	 * 
	 * @param	string		$errorMsg
	 * @param	string		$file
	 * @param	integer		$line
	 * @return	string				formatted error message
	 */
	public static function formatSyntaxError($errorMsg, $file = null, $line = null) {
		$errorMsg = 'Template compilation failed: '.$errorMsg;
		if ($file && $line) {
			$errorMsg .= " in template '$file' on line $line";
		}
		elseif ($file && !$line) {
			$errorMsg .= " in template '$file'";
		}
		return $errorMsg;
	}
	
	/**
	 * Compiles an {if} Tag
	 *
	 * @param 	string 		$tagArgs
	 * @param	boolean		$elseif		true, if this tag is an else tag
	 * @return	string 				php code of this tag
	 */
	protected function compileIfTag($tagArgs, $elseif = false) {
		// match valid if tag
		if (!preg_match('~^\s+!?'.$this->varWithModifierPattern.'(?:\s+(\|\||\&\&|>=?|<=?|===?|!==?)\s+!?'.$this->varWithModifierPattern.')*$~', $tagArgs)) {
			throw new SystemException($this->formatSyntaxError('syntax error in tag {'.($elseif ? 'elseif' : 'if').'}', $this->currentTemplate, $this->currentLineNo), 12000);
		}
		
		// find all variables
		$tagArgs = preg_replace('~(\s+!?)('.$this->varWithModifierPattern.')~e', '\'\\1\'.$this->compileVariableTag(StringUtil::unescape(\'\\2\'))', $tagArgs);
		
		return '<?php '.($elseif ? '} elseif' : 'if').' ('.$tagArgs.') { ?>';
	}
	
	/**
	 * Adds a tag to the tag stack.
	 *
	 * @param 	string 		$tag
	 */
	public function pushTag($tag) {
		$this->tagStack[] = array($tag, $this->currentLineNo);
	}
	
	/**
	 * Deletes a tag from the tag stack.
	 *
	 * @param 	string 		$tag
	 * @return 	string 		$tag
	 */
	public function popTag($tag) {
		list($openTag, $lineNo) = array_pop($this->tagStack);
		if ($tag == $openTag) {
			return $openTag;
		}
		if ($tag == 'if' && ($openTag == 'else' || $openTag == 'elseif')) {
			return $this->popTag($tag);
		}
	}
	
	/**
	 * Compiles an output tag.
	 * 
	 * @param	string		$tag
	 * @return	string			php code of this tag
	 */
	protected function compileOutputTag($tag) {
		$encodeHTML = false;
		$formatNumeric = false;
		if ($tag[0] == '@') {
			$tag = StringUtil::substring($tag, 1);
		}
		else if ($tag[0] == '#') {
			$tag = StringUtil::substring($tag, 1);
			$formatNumeric = true;
		}
		else {
			$encodeHTML = true;
		}
		
		$parsedTag = $this->compileVariableTag($tag);
		
		// the @ operator at the beginning of an output avoids
		// the default call of StringUtil::encodeHTML()
		if ($encodeHTML) {
			$parsedTag = 'StringUtil::encodeHTML('.$parsedTag.')';
		}
		// the # operator at the beginning of an output instructs
		// the complier to call the StringUtil::formatNumeric() method
		else if ($formatNumeric) {
			$parsedTag = 'StringUtil::formatNumeric('.$parsedTag.')';
		}
		
		return '<?php echo '.$parsedTag.'; ?>';
	}
	
	/**
	 * Escapes special characters . | [ $ : ->
	 * 
	 * @param	string		$string
	 * @return	string
	 */
	protected function escapeSpecialChars($string) {
		// escape specialchars
		$string = addcslashes($string, '.|[$:');
		$string = StringUtil::replace('->', '\->', $string);
		
		return $string;
	}
	
	/**
	 * Removes escapes from double quoted strings.
	 * 
	 * @param	string		$string
	 * @return	string
	 */
	public function handleDoubleQuotes($string) {
		// strip slashes from preg_replace eval
		$string = StringUtil::unescape($string);
		
		// escape $ in double quotes twice
		if (StringUtil::substring($string, 0, 1) == '"') {
			//$string = StringUtil::replace('$', '\$', $string);
			// parse unescaped simple vars in double quotes
			// replace $foo with {$this->v['foo']}
			$string = preg_replace('~'.$this->escapedPattern.$this->simpleVarPattern.'~', '{$this->v[\'\\1\']}', $string);
		}
		
		return $string;
	}
	
	/**
	 * Compiles a variable tag.
	 * 
	 * @param	string		$tag
	 * @return	string
	 */
	public function compileVariableTag($tag) {
		// escape syntax in quotes
		// replace "->.|[$:" with "\->\.\|\[\$\:" 
		$tag = preg_replace('~('.$this->quotePattern.')~e', '$this->escapeSpecialChars($this->handleDoubleQuotes(\'\\1\'))', $tag);
		
		// escape . (dot) in floats
		$tag = preg_replace('~((?:^|\(|\[|\s+|,|:|'.$this->mathOperatorsPattern.')\d+)(?=\.\d+)~', '\\1\\', $tag);
		
		// parse bean pattern
		// replace ->foo with ->getFoo()
		// $tag = preg_replace('~'.$this->escapedPattern.'->(\w+)(?![\(\w])~e', '\'->get\'.ucfirst(\'\\1\').\'()\'', $tag);
		
		// remove @ operator in object access
		// replace ->@foo with ->foo
		// $tag = preg_replace('~'.$this->escapedPattern.'->@~', '->', $tag);
		
		// parse dot syntax
		// replace .foo with [foo]
		$tag = preg_replace('~'.$this->escapedPattern.$this->dotAccessPattern.'~', '[\\1]', $tag);
		
		// parse modifiers
		// $foo|bar
		$this->modifiers = array();
		$tag = preg_replace('~'.$this->escapedPattern.$this->modifierPattern.'~e', '$this->parseModifier($this->handleDoubleQuotes(\'\\1\'))', $tag);
		foreach ($this->modifiers as $modifier) {
			// if necessary load plugin from plugin-dir
			if (file_exists($pluginFilename = $this->template->getPluginFilename('modifier', $modifier['name']))) {
				$className = 'TemplatePluginModifier'.StringUtil::firstCharToUpperCase(StringUtil::toLowerCase($modifier['name']));
				$this->autoloadPlugins[$className] = $pluginFilename;	
				
				array_unshift($modifier['parameters'], $tag);
				$tagArgs = $this->makeArgString($modifier['parameters']);
				$tag = "\\\$this->pluginObjects['".$className."']->execute(array(".$tagArgs."), \\\$this)";
			}
			else if (function_exists($modifier['name']) || in_array($modifier['name'], $this->unknownPHPFunctions)) {
				$parameters = '';
				foreach ($modifier['parameters'] as $parameter) {
					$parameters .= ','.$parameter;
				}
			
				$tag = $modifier['name'].'('.$tag.$parameters.')';	
			}
			else {
				throw new SystemException($this->formatSyntaxError("unknown modifier '".$modifier['name']."'", $this->currentTemplate, $this->currentLineNo), 12004);
			}
		}
		
		// parse bracket syntax
		// replace [foo] with ['foo']
		// do not touch [$foo]
		$tag = preg_replace('~'.$this->escapedPattern.'\[(\w+)\]~', '[\'\\1\']', $tag);
		
		// parse simple vars
		// replace $foo with $this->v['foo']
		$tag = preg_replace('~'.$this->escapedPattern.$this->simpleVarPattern.'~', '$this->v[\'\\1\']', $tag);
		
		// remove escapes in quotes
		$tag = preg_replace('~\\\\(?=->|\.|\||\[|\$|:)~', '', $tag);
		
		return $tag;
	}
	
	/**
	 * Parses a variable modifier.
	 * 
	 * @param	string		$modifier
	 * @return	string				an empty string
	 */
	protected function parseModifier($modifier) {
		$name = '';
		$parameters = $match = $matches = array();
		
		// find modifier name
		preg_match('~^\|(\w+)~', $modifier, $match);
		$name = $match[1];
		
		// parse modifier parameters
		preg_match_all('~'.$this->escapedPattern.':('.$this->varWithMathPattern.')~', $modifier, $matches);
		$parameters = $matches[1];
		
		$this->modifiers[] = array('name' => $name, 'parameters' => $parameters);
		
		return '';
	}
	
	/**
	 * Remove starting and ending quotes from the given string.
	 *
	 * @param 	string 		$string
	 * @return 	string
	 */
	public static function dequote($string) {
		if ((StringUtil::substring($string, 0, 1) == "'" || StringUtil::substring($string, 0, 1) == '"') && StringUtil::substring($string, -1) == StringUtil::substring($string, 0, 1)) {
			return StringUtil::substring($string, 1, -1);
		}
		else {
			return $string;
		}
	}
	
	/**
	 * Generates the regexp pattern.
	 */
	protected function buildPattern() {
		// negative lookbehind for a backslash
		$this->escapedPattern = '(?<!\\\\)';
		
		// valid variable name pattern
		$this->validVarnamePattern = '(?:[a-zA-Z_][a-zA-Z_0-9]*)';
		
		// matches constants
		// CONSTANT
		// __CONSTANT2
		$this->constantPattern = '(?:[A-Z_][A-Z_0-9]*)';
		
		// matches double quoted strings
		// "blah"
		// "quote: \"blah\""
		//$this->doubleQuotePattern = '"[^"\\\\]*(?:\\\\.[^"\\\\]*)*"';
		$this->doubleQuotePattern = '"(?:[^"\\\\]+|\\\\.)*"';
		
		// matches single quoted strings
		// 'blah'
		// 'don\'t'
		//$this->singleQuotePattern = '\'[^\'\\\\]*(?:\\\\.[^\'\\\\]*)*\'';
		$this->singleQuotePattern = '\'(?:[^\'\\\\]+|\\\\.)*\'';
		
		// matches single or double quoted strings
		$this->quotePattern = '(?:' . $this->doubleQuotePattern . '|' . $this->singleQuotePattern . ')';
		
		// matches numerical and boolean constants
		// 234
		// -12
		// 23.92
		// true
		// false
		$this->numericPattern = '(?i)(?:(?:\-?\d+(?:\.\d+)?)|true|false|null)';
		
		// matches simple variables
		// $foo
		$this->simpleVarPattern = '(?:\$('.$this->validVarnamePattern.'))';
		
		// matches array access (dot syntax)
		// .foo
		// .$foo
		// .12
		$this->dotAccessPattern = '(?:\.(\w+|'.$this->simpleVarPattern.'|'.$this->numericPattern.'|'.$this->quotePattern.'))';
		
		// matches array access (bracket syntax)
		// [foo]
		// [12]
		// [$foo]
		// [$foo.bar]
		$this->bracketAccessPattern = '(?:\[(?:\w+|'.$this->simpleVarPattern.'(?:'.$this->dotAccessPattern.')*|'.$this->numericPattern.'|'.$this->quotePattern.')\])';
		
		// matches all variables without object access
		$this->varWithoutObjectAccessPattern = '(?:'.$this->simpleVarPattern.'(?:'.$this->dotAccessPattern.'|'.$this->bracketAccessPattern.')*)';
		
		// matches object access (arrow syntax)
		// ->foo
		// ->@foo
		// ->foo()
		// ->foo($bar)
		// ->foo($bar, $foo)
		$this->objectMethodParamPattern = '(?:'.$this->varWithoutObjectAccessPattern.'|'.$this->quotePattern.'|'.$this->constantPattern.'|'.$this->numericPattern.')';
		$this->objectAccessPattern = '(?:->(?:\w+(?:\(\s*(?:'.$this->objectMethodParamPattern.'(?:\s*,\s*'.$this->objectMethodParamPattern.')*\s*)?\))(?![\.\[])|\w+))';
		
		// matches all variables
		$this->varPattern = '(?:'.$this->simpleVarPattern.'(?:'.$this->objectAccessPattern.'|'.$this->dotAccessPattern.'|'.$this->bracketAccessPattern.')*)';
		
		// matches variables, quotes, constants, numerics and math
		$this->varAndMorePattern = '(?:'.$this->constantPattern.'|'.$this->quotePattern.'|'.$this->numericPattern.'|'.$this->varPattern.')';
		
		// matches math operators
		// + - * / % ^
		//$this->mathOperatorsPattern = '[\+\-\*\/\%]';
		$this->mathOperatorsPattern = '\s*[\+\-\*\/\%\^]\s*';
		
		// matches variables with math operations
		$this->varWithMathPattern = '(?:'.$this->varAndMorePattern.'(?:'.$this->mathOperatorsPattern.$this->varAndMorePattern.')*)';
		
		// matches modifier
		// |foo
		// |foo:$bar
		$this->modifierPattern = '(\|\w+(?::'.$this->varWithMathPattern.')*)';
		
		// matches variable with modifier
		// $foo|bar
		$this->varWithModifierPattern = '(?:'.$this->varWithMathPattern.$this->modifierPattern.'*)';
		
		// matches variable outputs
		// @$oo
		$this->outputPattern = '(?:(?:@|#)?'.$this->varWithModifierPattern.')';
	}
	
	/**
	 * Returns the instance of the template class.
	 * 
	 * @return 	 Template
	 */
	public function getTemplate() {
		return $this->template;
	}
	
	/**
	 * Returns the left delimiter for template tags.
	 * 
	 * @return 	string
	 */
	public function getLeftDelimiter() {
		return $this->leftDelimiter;
	}
	
	/**
	 * Returns the right delimiter for template tags.
	 * 
	 * @return	string
	 */
	public function getRightDelimiter() {
		return $this->rightDelimiter;
	}
	
	/**
	 * Returns the name of the current template.
	 * 
	 * @return 	string
	 */
	public function getCurrentTemplate() {
		return $this->currentTemplate;
	}
	
	/**
	 * Returns the current line number.
	 * 
	 * @return 	integer
	 */
	public function getCurrentLineNo() {
		return $this->currentLineNo;
	}
}
?>